import json
import logging
from logging import getLogger

import yaml

from hakoniwa.entity.entity import Entity

from .context import Context
from .state import State


class Environment:
    """
    Environment is a domain defining the state machine each entity is wandering around.
    """

    def __init__(self, states: dict[State], context: Context) -> None:
        self.context = context
        self.states = states
        self.entities = []
        logging_handler = logging.FileHandler(self.context.history_file)
        self.logger = getLogger(__name__)
        self.logger.addHandler(logging_handler)
        self.iteration_count = 0

    @classmethod
    def from_yaml(cls, filename: str, context: Context = Context()):
        with open(filename, "r") as file:
            config = yaml.safe_load(file)
            states = {}
            for state_id, state in config["states"].items():
                states[state_id] = State(state_id, state["name"], state["choices"])
            context.need_interaction = config["context"]["need_interaction"]
            environment = cls(states, context)
            return environment

    def add_entity(self, entity):
        self.entities.append(entity)

    def next(self):
        for entity in self.entities:
            in_prompt = self._build_prompt(entity)
            out_response = entity.in_prompt(in_prompt)

            try:
                out_json = json.loads(out_response)
            except json.JSONDecodeError:
                self.logger.debug("Failed to parse response as JSON")
                return

            action = int(out_json["action"])
            choice = entity.state.choices[action]
            entity.to_state(self.states[choice["next"]])
            self._emit_log(entity.entity_id, choice, in_prompt)

        self.iteration_count += 1

    def _build_prompt(self, entity: Entity):
        state = entity.state
        choices = "\n"
        action_id = 0
        for _, choice in enumerate(state.choices):
            choices += f"  {action_id}: {choice['action']}\n"
            action_id += 1
        interections = ""
        for other in self.entities:
            if other.entity_id != entity.entity_id and other.state == state and self.context.need_interaction:
                self.logger.debug("Found other entity in the same state")
                interaction = other.interact(entity)
                interections += f"{other.entity_id} says '{interaction}'."
        prompt = f"""
        Context: You are in a state '{state.name}'. {interections}
        Actions:{choices}
        """

        self.logger.debug(prompt)

        return prompt

    def _build_interact_prompt(self, entity: Entity, other: Entity):
        prompt = f"""
        Context: You are in a state {entity.state} with {other.entity_id}. Do you want to say something to #{other.entity_id}?
        """

        return prompt

    def _emit_log(self, entity_id: str, choice: dict, prompt: str):
        record = {
            "iteration": self.iteration_count,
            "entity_id": entity_id,
            "action": choice["action"],
            "state": choice["next"],
            "prompt": prompt,
        }
        self.logger.info(json.dumps(record))



from logging import getLogger

import openai

from hakoniwa.entity.entity import Entity

logger = getLogger(__name__)


from .dynamic_entity import DynamicEntity


class OpenAIEntity(DynamicEntity):
    def __init__(self, entity_id, apikey, initial_state, personality=None):
        openai.api_key = apikey
        self.personality = personality
        super().__init__(entity_id=entity_id, initial_state=initial_state)

    def in_prompt(self, in_text=""):
        chat_completion = openai.ChatCompletion.create(
            model="gpt-3.5-turbo",
            messages=[
                {
                    "role": "system",
                    "content": f"{self.personality}"
                    + """
We expect you to take action each time based on the list of possible action and state you are now. You are given some input from the environment at each iteration align with the current state and action you took.
Please give us the ID of action you would take and the output text you would return to the environment according to the following JSON format which have an action ID you took in "action" field, and arbitrary output text in "output" field.

{
  "action": <Action ID>,
  "output": <Output Text>
}
""",
                },
                {"role": "user", "content": in_text},
            ],
        )

        if len(chat_completion.choices) == 0:
            return "{}"

        response = chat_completion.choices[0].message.content
        return response

    def interact(self, other: Entity):
        in_text = f"""{self.personality}".
        You are in a state '{self.state.name}' with {other.entity_id}. Say something to #{other.entity_id}?
        """
        chat_completion = openai.ChatCompletion.create(
            model="gpt-3.5-turbo",
            messages=[
                {
                    "role": "system",
                    "content": "Provide the response to the other entity.",
                },
                {"role": "user", "content": in_text},
            ],
        )

        if len(chat_completion.choices) == 0:
            return "{}"

        response = chat_completion.choices[0].message.content
        return response



from .entity import Entity


class DynamicEntity(Entity):
    """
    A DynamicEntity is an Entity that can create the response at runtime.
    """

    def __init__(self, entity_id, initial_state):
        super().__init__(entity_id=entity_id, initial_state=initial_state)

    def interact(self, in_text=""):
        pass



from hakoniwa.state import State


class Entity:
    def __init__(self, entity_id: str, initial_state: State) -> None:
        self.state = initial_state
        if entity_id is None:
            self.entity_id = id(self)
        else:
            self.entity_id = entity_id

    def in_prompt(self, in_text=""):
        return "FIXME"

    def to_state(self, state):
        self.state = state



from hakoniwa.entity.entity import Entity
from hakoniwa.state import State


class HelloEntity(Entity):
    """
    Entity always saying 'Hello World!'
    """

    def __init__(self, entity_id: str, initial_state: State) -> None:
        super().__init__(entity_id, initial_state)

    def in_prompt(self, in_text=""):
        return '{"action": 0, "output": "Hello World!"}'



from dataclasses import dataclass


@dataclass
class State:
    """
    State is a descrete field where an entity resides at the specific time of the iteration.
    """

    id: int
    name: str
    choices: list[dict]



import logging
import os
from argparse import ArgumentParser

import yaml

from hakoniwa.context import Context
from hakoniwa.entity import OpenAIEntity
from hakoniwa.environment import Environment


def build_environment(args):
    context = Context(args.history_file)
    environment = Environment.from_yaml(filename=args.env_file, context=context)
    with open(args.env_file, "r") as file:
        config = yaml.safe_load(file)
        for entity in config["entities"]:
            if entity["type"] == "openai":
                if "OPENAI_APIKEY" not in os.environ:
                    raise RuntimeError("OpenAI entity is not supported yet")
                environment.add_entity(
                    OpenAIEntity(
                        entity_id=entity["name"],
                        apikey=os.environ["OPENAI_APIKEY"],
                        initial_state=environment.states[entity["initial_state"]],
                        personality=entity["personality"],
                    )
                )
            else:
                raise RuntimeError("Unsupported entity type: {}".format(entity["type"]))

    return environment


def run_iterations(environment, max_iterations):
    for i in range(0, max_iterations):
        environment.next()


def run():
    parser = ArgumentParser(description="Hakoniwa CLI")
    parser.add_argument("-f", dest="env_file", help="File defining the environment")
    parser.add_argument("--iteration", dest="max_iterations", type=int, default=10, help="Max iterations")
    parser.add_argument("--history", dest="history_file", default="history.jsonl", help="History file")
    parser.add_argument("--debug", dest="debug", action="store_true", help="Enable debug log")

    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)
    environment = build_environment(args)
    run_iterations(environment, args.max_iterations)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hakoniwa', 'hakoniwa.entity']

package_data = \
{'': ['*']}

install_requires = \
['openai>=0.27.8,<0.28.0', 'torch>=2.0.1,<3.0.0', 'transformers>=4.31.0,<5.0.0']

entry_points = \
{'console_scripts': ['hakoniwa = hakoniwa.cli:run']}

setup_kwargs = {
    'name': 'hakoniwa',
    'version': '0.1.1',
    'description': '',
    'long_description': '# 箱庭 (Hakoniwa) ![Build](https://github.com/Lewuathe/hakoniwa/actions/workflows/main.yml/badge.svg)\n\n![hakoniwa](./hakoniwa.png)\n\n"Hakoniwa" which is a miniature garden in Japanese is a simulation framework letting LLM based entities play around inside. This framework aims to provide the way to experiment how the LLM behaves in the specific domain defined as a simple state machine which has states and actions respectively. The framework is designed to be able to be used for the following purposes:\n\n- To experiment how the LLM behaves in the specific enviroment.\n- To collect the behavior data resembling the human behavior in the real world.\n\n## Usage\n\n### Define the state machine\nFirst of all, we can define the state machine where each agent play around inside. The state machine is defined as a YAML file. The following is the example of the state machine definition.\n\n```yaml\nstates:\n  state0:\n    name: In the house\n    choices:\n      - action: Go outside\n        next: state1\n      - action: Stay inside\n        next: state0\n  state1:\n    name: Outside\n    choices:\n      - action: Go to the supermarket\n        next: state2\n      - action: Go back home\n        next: state0\n  state2:\n    name: In the supermarket\n    choices:\n      - action: Go back home\n        next: state0\n\nentities:\n  - name: A living thing\n    type: openai\n    initial_state: state0\n```\n\nThe state machine consists of the following two parts.\n- `states`: The list of states in the state machine. `choices` is the list of actions which the agent can take from the state.\n- `entities`: The list of entities which play around inside the state machine.\n\nThe CLI script, `hakoniwa` is usable to run the simulation with the state machine definition.\n\n```\n$ poetry run hakoniwa -f env.yaml\n```\n\nPlease make sure to set the `OPENAI_APIKEY` if you set the OpenAI entity in the environment. \n\nYou will see what it does in the environment.\n\n```\nINFO:hakoniwa.environment:A living thing,{\'action\': \'Stay inside\', \'next\': \'state0\'}\nINFO:hakoniwa.environment:A living thing,{\'action\': \'Go outside\', \'next\': \'state1\'}\nINFO:hakoniwa.environment:A living thing,{\'action\': \'Go to the supermarket\', \'next\': \'state2\'}\n```\n\n## Dependencies\n\nInstall [Poetry](https://python-poetry.org/) and run the following command to install dependencies.\n\n```\n$ poetry install\n```\n\n## Development\n\nWe can run the unit test with pytest.\n\n```\n$ make test\n```\n',
    'author': 'Kai Sasaki',
    'author_email': 'lewuathe@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
